function y = img_dcfree_get(x, dcfree, N, Tmin, Tmax)
% y = img_dcfree_get(x, dcfree, N, Tmin, Tmax)
% 
% This function constructs a dcfree image (y) from the input image (x). 
% dcfree is the dc free value being used and N is the block size.
% Default parameters: Tmin = 0, Tmax = 255, N = 8.
% If dcfree does not exist or is out of the valid range (such as -1), 
% the DC coefficient of each block will be assigned to be the middle point 
% of the valid range.
%
% Copyright (C) 2010 Junaid Jameel Ahmad, Shujun Li (www.hooklee.com)

if nargin<1
    disp('At least one input argument is needed!');
    return;
end

if ~exist('Tmax', 'var')
    Tmax = 255;
end
if ~exist('Tmin', 'var')
    Tmin = 0;
end
if Tmin>Tmax
    temp = Tmin;
    Tmin = Tmax;
    Tmax = temp;
end
if ~exist('N', 'var')
    N = 8;
end
dcfree_min = N * Tmin;
dcfree_max = N * Tmax;

% Converting x to double is very important to ensure the accuracy of the
% result! If x is an uint8 array, then all the following computations will
% be based on uint8 and the final result may be far from ideal.
if isinteger(x)
    y = double(x);
else
    y = x;
end

if (exist('dcfree', 'var') && isnumeric(dcfree) && ...
        dcfree>=dcfree_min && dcfree<=dcfree_max)
    % NO need to do computationally complex dct2 and idct2.
    % y_dct = dct2blockwise(y,N);
    % y_dct(1:N:end,1:N:end) = dcfree;
    % y = idct2blockwise(y_dct,N);
    for by=1:N:size(y,1)
        for bx=1:N:size(y,2)
            block = y(by:by+N-1,bx:bx+N-1);
            y(by:by+N-1,bx:bx+N-1) = block + (dcfree/N - mean(block(:)));
        end
    end
else
    [Y,X] = size(y);
    if (mod(Y,N)>0 || mod(X,N)>0)
        y = impad(y, N);
        [Y,X] = size(y);
    end
    for by=1:N:Y
        for bx=1:N:X
            block = y(by:by+N-1,bx:bx+N-1);
            range_min = Tmin-min(block(:));
            range_max = Tmax-max(block(:));
            y(by:by+N-1,bx:bx+N-1) = block + (range_min+range_max)/2;
        end
    end
end
