function img_dcfree = img_dcfree_update(img_dcfree, dcfree, scan, N, Tmin, Tmax)
% img_dcfree = img_dcfree_update(img_dcfree, dcfree, scan, N, Tmin, Tmax)
% This function is used to update the DC coefficient of the first block of
% a specific scan to be dcfree.
% Default values: Tmin = 0, Tmax = 255, N = 8, scan = 1,
% dcfree = the middle point of the valid range of the DC coefficient.
%
% Copyright (C) 2010 Shujun Li @ www.hooklee.com

if nargin<1
    disp('At least one input argument is needed!');
    return;
end

if ~exist('Tmax', 'var')
    Tmax = 255;
end
if ~exist('Tmin', 'var')
    Tmin = 0;
end
if Tmin>Tmax
    temp = Tmin;
    Tmin = Tmax;
    Tmax = temp;
end
if ~exist('N', 'var')
    N = 8;
end
if ~exist('scan', 'var')
    scan = 1;
end
dcfree_min = N * Tmin;
dcfree_max = N * Tmax;

% Converting x to double is very important as pointed out in
% img_dcfree_get function.
% Normally img_dcfree of this function is already double, but maybe there
% is possibility of having something wrong...
if isinteger(img_dcfree)
    img_dcfree = double(img_dcfree);
end

[Y X] = size(img_dcfree);
N_1 = N - 1;
switch scan
    case 1
        bx = 1:N;
        by = 1:N;
    case 2
        bx = X-N_1:X;
        by = 1:N;
    case 3
        bx = X-N_1:X;
        by = Y-N_1:Y;
    case 4
        bx = 1:N;
        by = Y-N_1:Y;
end

block = img_dcfree(by,bx);
if (exist('dcfree', 'var') && ~isempty(dcfree))
    if (ischar(dcfree) && strcmp(dcfree, 'AllBlocks'))
        % This is used to set the first DC based on the smallest range of
        % all blocks (as described in Uehara et al.'s paper).
        % This options is actually wrong and should not be used, but we
        % reserve it here for test purposes.
        delta_min = Tmin - Tmax;
        delta_max = Tmax - Tmin;
        for y=1:N:size(img_dcfree,1)
            for x=1:N:size(img_dcfree,2)
                block2 = img_dcfree(y:y+N_1,x:x+N_1);
                delta_min = max(delta_min, Tmin-min(block2(:)));
                delta_max = min(delta_max, Tmax-max(block2(:)));
            end
        end
        img_dcfree(by,bx) = block + (delta_min+delta_max)/2;
    elseif (dcfree>=dcfree_min && dcfree<=dcfree_max)
        % NO need to do computationally complex dct2 and idct2.
        % block_dct2 = dct2(block);
        % block_dct2(1,1) = dcfree;
        % img_dcfree(by,bx) = idct2(block_dct2);
        img_dcfree(by,bx) = img_dcfree(by,bx) + (dcfree/N - mean(block(:)));
    else % Invalid dcfree = use default
        range_min = Tmin-min(block(:));
        range_max = Tmax-max(block(:));
        img_dcfree(by,bx) = block + (range_min+range_max)/2;
    end
else % Default = Middle point of valid range
    range_min = Tmin-min(block(:));
    range_max = Tmax-max(block(:));
    img_dcfree(by,bx) = block + (range_min+range_max)/2;
end
