function [img_recovered, dcfree_scans, img, img_dcfree, img_scans] = imrecover_FRM(arg1, options, hWaitbar)
% [img_recovered, dcfree_scans, img, img_dcfree, img_scans] = imrecover_FRM(img_filepath, options, hWaitbar)
% [img_recovered, dcfree_scans, img, img_dcfree, img_scans] = imrecover_FRM(img_dcfree, options, hWaitbar)
% 
% This function recovers an image by minimizing the underflow/overflow rate
% of one or multiple scans and then averaging the obtained results.
% The minimal underflow/overflow rate is determined by a halving method to
% speedup the searching process.
% Here, "FRM" denotes "flow rate minimization".
% 
% If the XML toolbox is available, it will also create an XML file to store
% the best dcfree values for all the scans.
% Different options will have different entries in the XML document.
% 
% The input argumen options is a structure variable composed of the
% following fields:
% Tmin, Tmax: minimal/maximal pixel value
% blocksize: size of block
% scan: a vector of one to four scans
% bound_dc, weighted_block: see help of imrecover_scan function
% use_USO_range: employ Uehara et al.'s method to determine the DC range of
%                the first reference block
% dcfree_compensation: compensation for the dcfree calculated by minimizing
%                      the flow rate
% postprocess: do histogram shift and clipping on the recovered image
% PSNR_max: the expected PSNR of the first reference block, which denotes
%           the accuracy of the first unknown DC coefficient (the actual
%           PSNR of the whole image is normally smaller than this value,
%           so this parameter should be set larger than the expected PSNR
%           of the whole image)
% flowrate_precision: precision of flow rate, a precentage
% Default values: Tmin = 0, Tmax = 255, blocksize = 8, scans = 1:4, 
% weighted_block = false, dcfree_compensation = 0, postprocess = false,
% PSNR_max = 20*log10((Tmax-Tmin)/1), flowrate_precision = 0.1%.
%
% The optional argument "hWaitbar" is used to update the status of a
% waitbar (for demo purpose only).
%
% Copyright (C) 2010 Shujun Li @ www.hooklee.com

img_recovered = [];
dcfree_scans = [];
img = [];
img_dcfree = [];
img_scans = [];

if nargin<1
    disp('At least one input argument is needed!');
    return;
end

if ~exist('hWaitbar','var')
    hWaitbar = [];
end
if (~exist('options','var') || ~isstruct(options))
    options = []; % Supress warning for ~isstruct(options).
    options.Tmin = 0;
    options.Tmax = 255;
    options.blocksize = 8;
    options.scan = 1:4;
    options.bound_dc = true;
    options.weighted_block = false;
    options.weighted_scans = false;
    options.dcfree_compensation = 0;
    options.use_USO_range = false;
    options.postprocess = false;
    options.PSNR_max = 20*log10((options.Tmax-options.Tmin)/1); % 48.1308 dB for [0 255]
    options.flowrate_precision = 0.001; % 0.1%
    options.FR_type = 1;
else
    % Check if each filed is included in options.
    options_fieldnames = {'Tmin' 'Tmax' 'blocksize' 'scan' 'bound_dc' 'weighted_block' ...
        'weighted_scans' 'dcfree_compensation' 'use_USO_range' 'postprocess' 'PSNR_max' 'flowrate_precision' 'FR_type'};
    for i=1:numel(options_fieldnames)
        if ~isfield(options, options_fieldnames{i})
            switch options_fieldnames{i}
                case 'Tmin'
                    options.Tmin = 0;
                case 'Tmax'
                    options.Tmax = 255;
                case 'blocksize'
                    options.blocksize = 8;
                case 'scan'
                    options.scan = 1:4;
                case 'bound_dc'
                    options.bound_dc = true;
                case 'weighted_block'
                    options.weighted_block = false;
                case 'weighted_scans'
                    options.weighted_scans = false;
                case 'dcfree_compensation'
                    options.dcfree_compensation = 0;
                case 'use_USO_range'
                    options.use_USO_range = false;
                case 'postprocess'
                    options.postprocess = false;
                case 'PSNR_max'
                    options.PSNR_max = 20*log10((options.Tmax-options.Tmin)/1); % 48.1308 dB for [0 255]
                case 'flowrate_precision'
                    options.flowrate_precision = 0.001;
                case 'FR_type'
                    options.FR_type = 1;
            end
        end
    end
end
Tmin = options.Tmin;
Tmax = options.Tmax;
N = options.blocksize;
% When the postprocess is true, it means that the average image recovered
% from the four scans should be postprocessed.
% Then, it should be more beneficial if we do NOT do the postprocess for
% the four scans.
postprocess = options.postprocess;
if postprocess
    options.postprocess = false;
end
% Remove invalid scan numbers.
options.scan = options.scan(options.scan>=1 & options.scan<=4);
if isempty(options.scan)
    disp('The "scan" field of the input argument options is invalid!');
    return;
end
scan_number = numel(options.scan);
% When there is only one scan, the advanced weighting scheme should not be
% used; otherwise 1) 0/0=Inf will happen; 2) the process will become slow.
if (scan_number==1 && options.weighted_scans)
    scan.weighted_scans = false;
end

dcfree_scans = cell(1,4);
if ischar(arg1) % img_filename
    try
        img_filepath = arg1;
        fprintf('Loading image from "%s"...\n', arg1);
        img = imread(arg1);
        img = impad(img, N);
        img_dcfree = img_dcfree_get(img, -1, N, Tmin, Tmax);
        dcfree_scans = xmlread4FRM(img_filepath, options);
        % Calculate the real dcfree value from non-compensated one.
        if options.dcfree_compensation~=0
            for i=1:4
                if isempty(dcfree_scans{i})
                    continue;
                end
                switch i
                    case 1
                        block = img_dcfree(1:N,1:N);
                    case 2
                        block = img_dcfree(1:N,end-7:end);
                    case 3
                        block = img_dcfree(end-7:end,end-7:end);
                    case 4
                        block = img_dcfree(end-7:end,1:N);
                end
                dcfree_min = N * double(mean(block(:))+Tmin-min(block(:)));
                dcfree_max = N * double(mean(block(:))+Tmax-max(block(:)));
                if options.dcfree_compensation<0
                    dcfree_scans{i} = max(dcfree_min, dcfree_scans{i}+options.dcfree_compensation);
                elseif options.dcfree_compensation>0
                    dcfree_scans{i} = min(dcfree_max, dcfree_scans{i}+options.dcfree_compensation);
                end
            end
        end
    catch me
        fprintf('Error: %s', me.message);
        return;
    end
else
    img_filepath = [];
    img_dcfree = arg1; % This should already be a DC-free image.
end

img_scans = {0 0 0 0};
scan = options.scan; % NOTE: options.scan will be overwritten inside the loop!
update_XML = false;
t0 = tic;
for index=1:scan_number
    i = scan(index);
    fprintf('=============Scan #%d================\n', i);
    options.scan = i;
    if isempty(dcfree_scans{i})
        [img_scans{i}, dcfree_scans{i}] = imrecover_scan_FRM(img_dcfree, options, hWaitbar, (index-1)/scan_number, t0);
        update_XML = true;
    else
        fprintf('Calculating the recovered image based on the dcfree value (%g) loaded from the XML document...\n', dcfree_scans{i});
        options.dcfree = dcfree_scans{i};
        img_scans{i} = imrecover_scan(img_dcfree, options);
    end
    if ~isempty(hWaitbar)
        waitbar(index/scan_number, hWaitbar, sprintf('Please wait, the process may last a few minutes...\nTime elapsed: %g seconds', toc(t0)));
    end
end
toc(t0);
disp('Generating the final recovered image from scans...');
if options.weighted_scans
    img_recovered = weighted_scans(img_scans, scan, 0);
else
    img_recovered = weighted_scans(img_scans, scan, 1);
end
if postprocess
    disp('Postprocessing the recovered image...');
    img_recovered = shift2center(img_recovered, 1, Tmin, Tmax);
    img_recovered = round(clip(img_recovered, Tmin, Tmax)); % Rounding after clipping may help increase PSNR.
end

if (~isempty(img_filepath) && update_XML && ...
        exist('xml_load.m','file')==2 && exist('xml_save.m','file')==2)
    disp('Saving/Updating data in XML document...');
    options.scan = scan; % Restore the scan field.
    xmlwrite4FRM(img_filepath, dcfree_scans, options);
end


function [img_recovered, dcfree, flow_rate] = imrecover_scan_FRM(img_dcfree, options, hWaitbar, percentage, t0)
% [img_recovered, dcfree, flow_rate] = imrecover_scan_FRM(img_dcfree, options)
% This internal function recovers an image by minimizing the
% underflow/overflow rate for a specific scan.

% Remove this field to avoid potential problems.
if isfield(options, 'dcfree')
    options = rmfield(options, 'dcfree');
end

img_recovered = [];
dcfree = [];
flow_rate = [];

Tmin = options.Tmin;
Tmax = options.Tmax;
N = options.blocksize;
dc_min = Tmin * N;
dc_max = Tmax * N;
dc_mid = (dc_min+dc_max)/2;
dcfree_precision = (Tmax-Tmin)/10^(options.PSNR_max/20) * N;

block_number = numel(img_dcfree)/N^2;

switch options.scan
    case 1
        block = img_dcfree(1:N,1:N);
    case 2
        block = img_dcfree(1:N,end-7:end);
    case 3
        block = img_dcfree(end-7:end,end-7:end);
    case 4
        block = img_dcfree(end-7:end,1:N);
end
dcfree_min = N * double(mean(block(:))+Tmin-min(block(:)));
dcfree_max = N * double(mean(block(:))+Tmax-max(block(:)));
% Try if use_USO_range helps promote the performance.
if options.use_USO_range
    img_USO = imrecover_scan(img_dcfree_update(img_dcfree,dc_mid,options.scan,N,Tmin,Tmax), options);
    [img_USO, range_USO] = bound_DC_image(img_USO, N, Tmin, Tmax);
    range_USO = range_USO + dc_mid;
    if (range_USO(1)<=range_USO(2) && range_USO(2)-range_USO(1)<dcfree_max-dcfree_min)
        dcfree_min = range_USO(1);
        dcfree_max = range_USO(2);
    end
    if ~isempty(hWaitbar)
        waitbar(percentage, hWaitbar, sprintf('Please wait, the process may last a few minutes...\nTime elapsed: %g seconds', toc(t0)));
    end
end
dcfree_mid = (dcfree_min+dcfree_max)/2;
if dcfree_min>=dcfree_max % When there is only one possible dcfree value.
    dcfrees = dcfree_mid;
else
    % The first set of dcfree values are sampled with an interval of W/4, where
    % W is the length of the valid range of dcfree values.
    dcfrees = [dcfree_min (dcfree_min+dcfree_mid)/2 dcfree_mid (dcfree_mid+dcfree_max)/2 dcfree_max];
end

flow_rates = 2*ones(size(dcfrees));
imgs_recovered = cell(size(dcfrees));
imrecover_scan_counts = 0;
disp('Looking for the minimal underflow/overflow rate to recover the original image...');
while 1
    for i=1:numel(dcfrees)
        if flow_rates(i)>1 % If new sampled dcfree values without known flow rates.
            img_dcfree = img_dcfree_update(img_dcfree, dcfrees(i), options.scan, N, Tmin, Tmax);
            [imgs_recovered{i}, flow_rates(i)] = imrecover_scan(img_dcfree, options);
            % flow_rates(i) = flow_rates(i)/block_number; % No need to make it a percentage.
            imrecover_scan_counts = imrecover_scan_counts + 1;
            fprintf('dcfree = %g: \tflow rate = %g\n', dcfrees(i), flow_rates(i));
            if ~isempty(hWaitbar)
                waitbar(percentage, hWaitbar, sprintf('Please wait, the process may last a few minutes...\nTime elapsed: %g seconds', toc(t0)));
            end
        end
    end
    flow_rates = round(flow_rates/options.flowrate_precision)*options.flowrate_precision;
    flow_rate_min = min(flow_rates);
    indices_flow_rate_min = find(flow_rates==flow_rate_min);
    % We choose the dcfree value closest to the middle point of the
    % region with the minimal flow.
    dcfree_flow_rate_min = mean(dcfrees(indices_flow_rate_min([1 end])));
    index = find(dcfrees==dcfree_flow_rate_min);
    img_recovered_renewed = false;
    if numel(index)>1
        disp('Error: find(dcfrees==dcfree_temp) returns more than one indices, which should never happen! Check the code for bugs!');
        return;
    elseif numel(index)==1
        dcfree = dcfrees(index);
        flow_rate = flow_rates(index);
        img_recovered = imgs_recovered{index};
    else % numel(index)==0
        % min(dcfrees)<dcfree_temp<max(dcfrees) => 1<=index<numel(dcfrees).
        dcfree = dcfree_flow_rate_min;
        img_recovered_renewed = true;
    end
    % If the precision of the dcfree value has been reached, the algorithm
    % stops.
    if (numel(dcfrees)==1 || dcfrees(2)-dcfrees(1)<=dcfree_precision)
        if (img_recovered_renewed || options.dcfree_compensation~=0)
            % Compensated dcfree should not be out of the valid range.
            dcfree_old = dcfree;
            if options.dcfree_compensation<0
                dcfree = max(dcfree_min, dcfree+options.dcfree_compensation);
            elseif options.dcfree_compensation>0
                dcfree = min(dcfree_max, dcfree+options.dcfree_compensation);
            end
            if (img_recovered_renewed || dcfree_old~=dcfree)
                img_dcfree = img_dcfree_update(img_dcfree, dcfree, options.scan, N, Tmin, Tmax);
                [img_recovered, flow_rate] = imrecover_scan(img_dcfree, options);
                flow_rate = flow_rate/block_number;
                imrecover_scan_counts = imrecover_scan_counts + 1;
                if ~isempty(hWaitbar)
                    waitbar(percentage, hWaitbar, sprintf('Please wait, the process may last a few minutes...\nTime elapsed: %g seconds', toc(t0)));
                end
            end
            if img_recovered_renewed
                fprintf('dcfree (averaged) = %g: \tflow rate = %g%%\n', dcfree, 100*flow_rate);
            else
                fprintf('dcfree (compensated) = %g: \tflow rate = %g%%\n', dcfree, 100*flow_rate);
            end
        end
        break;
    end
    % Find the index of the left boundary of the new search window.
    index_left = 1;
    for i=indices_flow_rate_min(1)-1:-1:2
        if flow_rates(i)>flow_rate_min
            index_left = i;
            break;
        end
    end
    % Find the index of the right boundary of the new search window.
    index_right = numel(dcfrees);
    for i=indices_flow_rate_min(end)+1:numel(dcfrees)-1
        if flow_rates(i)>flow_rate_min
            index_right = i;
            break;
        end
    end
    % Remove elements out of the new search window.
    % In the ideal case, only 3 elements will be left.
    % In the worst case, all the elements are left.
    indices_removed = [1:index_left-1 index_right+1:numel(dcfrees)];
    dcfrees(indices_removed) = [];
    flow_rates(indices_removed) = [];
    imgs_recovered(indices_removed) = [];
    % Extend the three vectors to add numel(dcfrees)-1 new elements, which
    % are inserted between pairs of adjacent old elements.
    dcfrees = [dcfrees(1:end);(dcfrees(1:end-1)+dcfrees(2:end))/2 0];
    dcfrees = dcfrees(:)';
    dcfrees(end) = [];
    flow_rates = [flow_rates(1:end);2*ones(size(flow_rates))];
    flow_rates = flow_rates(:)';
    flow_rates(end) = [];
    imgs_recovered{2,1} = [];
    imgs_recovered = imgs_recovered(:)';
    imgs_recovered(end) = [];
end

if options.postprocess
    disp('Postprocessing the recovered image...');
    img_recovered = shift2center(img_recovered, 1, Tmin, Tmax);
    img_recovered = clip(img_recovered, Tmin, Tmax);
elseif ~options.bound_dc % In this case, at least the output should be clipped.
    img_recovered = clip(img_recovered, Tmin, Tmax);
end

disp('The whole process has been finished!');
fprintf('Number of calls of imrecover_scan function: %d\n', imrecover_scan_counts);
fprintf('dcfree (final) = %g: \tflow rate = %g%%\n', dcfree, 100*flow_rate);


function img_recovered = weighted_scans(img_scans, scan, a)
% This function generates img_recovered by using different weights for each
% pixel in the four scans.

if nargin<1
    disp('At least one input argument is needed!');
    return;
end

if ~exist('scan','var')
    scan = 1:4;
end

if ~exist('a','var')
    a = 0; % a is the weight of the point furthest to the reference block.
end
if a==1 % This means fixed weight for all pixels -- naive averaging.
    img_recovered = (img_scans{1}+img_scans{2}+img_scans{3}+img_scans{4})/numel(scan);
    return;
end

[img_Y, img_X] = size(img_scans{1});
corners = [1 1;1 img_X;img_Y img_X;img_Y 1];
d_max = sqrt((img_X-1)^2+(img_Y-1)^2);
a = -(1-a)/d_max;
b = 1; % b is the weight of the point closest to the the reference block.
Xs = repmat(1:img_X, img_Y, 1);
Ys = repmat((1:img_Y)', 1, img_X);
W_sum = zeros(img_Y, img_X);
img_recovered = zeros(img_Y, img_X);
for i=1:numel(scan)
    W = a*sqrt((Xs-corners(scan(i),2)).^2+(Ys-corners(scan(i),1)).^2) + b;
    W_sum = W_sum + W;
    img_recovered = img_recovered + W.*img_scans{scan(i)};
end
img_recovered = img_recovered./W_sum;


function dcfrees = xmlread4FRM(img_filepath, options, scans)
% dcfrees = xmlread4FRM(img_filepath, options, scans)
% This function reads the dcfree values stored in an XML document, which
% were earlier calculated by imrecover_FRM function.
% The third argument scans is a vector to selectively output some of the
% dcfree values of the four scans.
% The returned value dcfrees is a 1X4 cell array to hold the four dcfree
% values.
% When there is error or no match of options is found, an empty array is
% returned.

dcfrees =cell(1,4);

if exist('xml_load.m','file')~=2
    return;
end

if ~exist('img_filepath','var')
    disp('At least one input argument is needed!');
    return;
end

if (~exist('scans','var') || isempty(scans))
    scans = 1:4;
end
% If options does not exist, use the default value of imrecover_FRM
% function.
if ~exist('options', 'var')
    options.Tmin = 0;
    options.Tmax = 255;
    options.blocksize = 8;
    options.bound_dc = true;
    options.weighted_block = false;
    options.use_USO_range = false;
    options.PSNR_max = 20*log10((options.Tmax-options.Tmin)/1); % 48.1308 dB for [0 255]
    options.flowrate_precision = 0.001; % 0.1%
    options.FR_type = 1;
else
    % Check if each filed is included in options.
    % postprocess does not influence the dcfree values, so it should be
    % excluded from the fields.
    options_fieldnames = {'Tmin' 'Tmax' 'blocksize' 'bound_dc' 'weighted_block' ...
        'use_USO_range' 'PSNR_max' 'flowrate_precision' 'FR_type'};
    for i=1:numel(options_fieldnames)
        if ~isfield(options, options_fieldnames{i})
            switch options_fieldnames{i}
                case 'Tmin'
                    options.Tmin = 0;
                case 'Tmax'
                    options.Tmax = 255;
                case 'blocksize'
                    options.blocksize = 8;
                case 'bound_dc'
                    options.bound_dc = true;
                case 'weighted_block'
                    options.weighted_block = false;
                case 'use_USO_range'
                    options.use_USO_range = false;
                case 'PSNR_max'
                    options.PSNR_max = 20*log10((options.Tmax-options.Tmin)/1); % 48.1308 dB for [0 255]
                case 'flowrate_precision'
                    options.flowrate_precision = 0.001;
                case 'FR_type'
                    options.FR_type = 1;
            end
        end
    end
    % Remove unnecessary fields.
    options_fieldnames2 = fieldnames(options);
    for i=1:numel(options_fieldnames2)
        if ~ismember(options_fieldnames2{i}, options_fieldnames)
            options = rmfield(options, options_fieldnames2{i});
        end
    end
end
% Saving data in an XML document causes precision loss, so we should set a
% PSNR precision to make data comparison possible if PSNR_max is calculated
% from the equation 20*log10((Tmax-Tmin)/pixel_precision)).
PSNR_precision = 0.001;
options.PSNR_max = round(options.PSNR_max/PSNR_precision)*PSNR_precision;

[img_filefolder, img_filetitle, ext, versn] = fileparts(img_filepath);
img_filename = [img_filetitle ext versn];
xml_filepath = fullfile(img_filefolder, [img_filetitle '.xml']);

if exist(xml_filepath, 'file')~=2
    return;
end

v = xml_load(xml_filepath);
try
    if ~strcmp(v.name, img_filename)
        return;
    end
    options_number = numel(v.options);
    for i=1:options_number
        options_i = v.options{i};
        options_i = rmfield(options_i, 'dcfree_FRM');
        if isfield(options_i, 'dcfree_compensation')
            if options_i.dcfree_compensation~=0
                continue; % This field may appear in old XML document, which should be simply skipped.
            else
                options_i = rmfield(options_i, 'dcfree_compensation');
            end
        end
        options_i.PSNR_max = round(options_i.PSNR_max/PSNR_precision)*PSNR_precision;
        if isequal(options, options_i)
            scans = scans(scans>=1 & scans<=4);
            dcfrees = v.options{i}.dcfree_FRM(scans);
            return;
        end
    end
    % Return {[],[],[],[]} if the current options is not in the XML file.
catch me
    fprintf('Error: %s\n', me.message);
    return;
end


function xmlwrite4FRM(img_filepath, dcfrees, options)
% xmlwrite4FRM(img_filepath, dcfrees, options)
% Write dcfree values corresponding to the four FRM scans and related 
% options into an XML document for future reading.
% This function will help avoid repeated invokations of imrecover_FRM
% functions, which are very time consuming for a large number of images.
% Different options will store as different items of "options" node in the
% XML document.
%
% The second argument dcfrees must be a 4-tuple cell array, which 
% represents the first to the fourth scans.
% If it is not a cell array, the function will convert it to a cell array.
% In case there are less than 4 elements, use [] to pad it.
% 
% This function depends on a third-party XML toolbox available at the
% following link: http://www.geodise.org/toolboxes/generic/xml_toolbox.htm

if (exist('xml_load.m','file')~=2 || exist('xml_save.m','file')~=2)
    return;
end

if ~exist('img_filepath','var')
    disp('At least one input argument is needed!');
    return;
end

% If options does not exist, use the default value of imrecover_FRM
% function.
if (~exist('options', 'var') || ~isstruct(options))
    options = []; % Supress warnings when ~isstruct(options)
    options.Tmin = 0;
    options.Tmax = 255;
    options.blocksize = 8;
    options.bound_dc = true;
    options.weighted_block = false;
    options.use_USO_range = false;
    options.PSNR_max = 20*log10((options.Tmax-options.Tmin)/1); % 48.1308 dB for [0 255]
    options.flowrate_precision = 0.001; % 0.1%
    options.FR_type = 1;
else
    % Check if each filed is included in options.
    % postprocess does not influence the dcfree values, so it should be
    % excluded from the fields.
    options_fieldnames = {'Tmin' 'Tmax' 'blocksize' 'bound_dc' 'weighted_block' ...
        'use_USO_range' 'PSNR_max' 'flowrate_precision' 'FR_type'};
    for i=1:numel(options_fieldnames)
        if ~isfield(options, options_fieldnames{i})
            switch options_fieldnames{i}
                case 'Tmin'
                    options.Tmin = 0;
                case 'Tmax'
                    options.Tmax = 255;
                case 'blocksize'
                    options.blocksize = 8;
                case 'bound_dc'
                    options.bound_dc = true;
                case 'weighted_block'
                    options.weighted_block = false;
                case 'use_USO_range'
                    options.use_USO_range = false;
                case 'PSNR_max'
                    options.PSNR_max = 20*log10((options.Tmax-options.Tmin)/1); % 48.1308 dB for [0 255]
                case 'flowrate_precision'
                    options.flowrate_precision = 0.001;
                case 'FR_type'
                    options.FR_type = 1;
            end
        end
    end
    % Remove unnecessary fields.
    options_fieldnames2 = fieldnames(options);
    for i=1:numel(options_fieldnames2)
        if ~ismember(options_fieldnames2{i}, options_fieldnames)
            options = rmfield(options, options_fieldnames2{i});
        end
    end
end
% Saving data in an XML document causes precision loss, so we should set a
% PSNR precision to make data comparison possible if PSNR_max is calculated
% from the equation 20*log10((Tmax-Tmin)/pixel_precision)).
PSNR_precision = 0.001;
options.PSNR_max = round(options.PSNR_max/PSNR_precision)*PSNR_precision;

if ~exist('dcfrees','var')
    dcfrees = cell(1,4); % Initialize the xml document.
elseif ~iscell(dcfrees)
    dcfrees = num2cell(dcfrees);
end
if numel(dcfrees)<2
    dcfrees = cat(dcfrees, cell(1,4-numel(dcfrees)));
end

[img_filefolder, img_filetitle, ext, versn] = fileparts(img_filepath);
img_filename = [img_filetitle ext versn];
xml_filepath = fullfile(img_filefolder, [img_filetitle '.xml']);
options_index = 0;
if exist(xml_filepath, 'file')==2
    v = xml_load(xml_filepath);
    try
        if ~strcmp(v.name, img_filename)
            disp('Error: This XML file is not for this image!');
            return;
        end
        options_number = numel(v.options);
        for i=1:options_number
            options_i = v.options{i};
            options_i = rmfield(options_i, 'dcfree_FRM');
            if isfield(options_i, 'dcfree_compensation') % This field may appear in old XML document.
                if options_i.dcfree_compensation~=0
                    continue; % This field may appear in old XML document, which should be simply skipped.
                else
                    options_i = rmfield(options_i, 'dcfree_compensation');
                end
            end
            options_i.PSNR_max = round(options_i.PSNR_max/PSNR_precision)*PSNR_precision;
            if isequal(options, options_i)
                options_index = i;
                break;
            end
        end
        if options_index==0 % The current options is not in the XML file.
            options_index = options_number + 1;
            v.options{options_index} = options;
        end
    catch me
        fprintf('Error: %s\n', me.message);
        return;
    end
else
    v.name = img_filename;
    options_index = 1;
    v.options{options_index} = options;
end

v.options{options_index}.dcfree_FRM = dcfrees;

xml_save(xml_filepath, v);
