function [img_recovered, img_scans] = imrecover_USO(img_dcfree, options)
% [img_recovered, img_scans] = imrecover_USO(img_dcfree, options)
% This function implements the basic DC recovery method reported in the
% Uehara et al.'s paper [IEEE Trans. IP, 15(11):3592, 2006].
%
% Since Ueharat et al.'s paper does not explain all the details very
% clearly, there are several different interpretations of how the bounding
% dc function should be done:
% 1) bound the DC value of each block DURING the scanning process, i.e.,
%    Step II-E2 is done inside each scan of Step II-E1.
% 2) bound the DC values of all blocks AFTER each scanning process, i.e.,
%    Step II-E2 is done after each scan in Step II-E1 finishes.
% 3) bound the DC values of all blocks AFTER all scanning processes, i.e.,
%    Step II-E2 is done after all scans in Step II-E1 finish.
% Although Interpretation 1 seems to be the most reasonable one, some
% sentences in Uehara et al.'s paper seem to imply Interpretation 2 or 3.
% A careful read of Section II-E2 led us to believe that Interpretation 2
% is the correct understanding.
% This function is implemented to have default values for Interpretation 2, 
% but it also offers switches for testing other interpretations.
%
% It is also not clear how the bounding should be done. It seems that only
% the first reference block's DC value and the average intensity of the
% whole image will be adjusted.
% 
% For pixel values out of the valid range after the DC-bounding process, 
% Uehara et al.'s paper says "either all pixel values must be scaled or 
% only the pixel values outside the valid range must be adjusted."
% Unfornately, nothing is said which one is used in their implementation
% and which kind of adjustments can be done for the second option.
% In this function, we offer two possible solutions: re-scaling and
% clipping. The former is the default since Uehara et al.'s paper mentiones
% it first.
%
% Copyright (C) 2010 Shujun Li @ www.hooklee.com

img_recovered = [];
img_scans = [];

if nargin<1
    disp('At least one input argument is needed!');
    return;
end

if (~exist('options', 'var') || ~isstruct(options))
    options = []; % Supress warnings when ~isstruct(options)
    options.Tmin = 0;
    options.Tmax = 255;
    options.blocksize = 8;
    options.bound_dc = false; % bound_dc used inside imrecover_scan function.
    options.weighted_block = false;
    options.bound_dc_scans = true; % bound_dc applied to each image returned by imrecover_scan.
    options.bound_dc_global = false; % bound_dc applied to the averaged image from four scans.
    options.postprocess_final = 'adjust_image';
else
    % Check if each field is included in options.
    options_fieldnames = {'Tmin' 'Tmax' 'blocksize' 'bound_dc' 'weighted_block' ...
        'bound_dc_scans' 'bound_dc_global' 'postprocess_final'};
    for i=1:numel(options_fieldnames)
        if ~isfield(options, options_fieldnames{i})
            switch options_fieldnames{i}
                case 'Tmin'
                    options.Tmin = 0;
                case 'Tmax'
                    options.Tmax = 255;
                case 'blocksize'
                    options.blocksize = 8;
                case 'bound_dc'
                    options.bound_dc = false;
                case 'weighted_block'
                    options.weighted_block = false;
                case 'bound_dc_scans'
                    options.bound_dc_scans = true;
                case 'bound_dc_global'
                    options.bound_dc_global = false;
                case 'postprocess_final'
                    options.postprocess_final = 'adjust_image';
            end
        end
    end
end
N = options.blocksize;
Tmin = options.Tmin;
Tmax = options.Tmax;

% Get the four scan images.
img_scans = cell(1,4);
for s=1:4
    options.scan = s;
    img_scans{s} = imrecover_scan(img_dcfree, options);
    if options.bound_dc_scans
        img_scans{s} = bound_DC_image(img_scans{s}, N, Tmin, Tmax);
    end
end

% Calculate the final recovered image from the four scan images.
img_recovered = (img_scans{1}+img_scans{2}+img_scans{3}+img_scans{4})/4;

% Step E-II: bounding the DC values.
if options.bound_dc_global
    img_recovered = bound_DC_image(img_recovered, N, Tmin, Tmax);
end

switch options.postprocess_final
    case {'adjust' 'adjust_image' 'scale_OR_shift' 'scale OR shift'}
        img_recovered = im_scale_OR_shift(img_recovered, Tmin, Tmax);
    case 'adjust_blocks' % Blockwise im_blocks_scale_OR_shift
        img_recovered = blocks_scale_OR_shift(img_recovered, N, Tmin, Tmax);
    case {'clip' 'clipping'}
        img_recovered = clip(img_recovered, Tmin, Tmax);
    otherwise % do nothing
end


function img = blocks_scale_OR_shift(img, N, Tmin, Tmax)
% img = blocks_scale_OR_shift(img, N, Tmin, Tmax)
% Adjust blocks whose pixel values are out of range.
% Two possible adjustments are made according to the range size of the
% pixel values:
% 1) if the range size is larger than Tmax-Tmin, do rescaling;
% 2) if the range size is smaller than Tmax-Tmin, shift the whole block
%    towards the valid range until no pixel values are out of range.

if ~exist('Tmax', 'var')
    Tmax = 255;
end
if ~exist('Tmin', 'var')
    Tmin = 0;
end
if Tmin>Tmax
    temp = Tmin;
    Tmin = Tmax;
    Tmax = temp;
end
if ~exist('N', 'var')
    N = 8;
end

[Y,X] = size(img);
Twidth = Tmax - Tmin;
for y=1:N:Y
    for x=1:N:X
        block = img(y:y+N-1,x:x+N-1);
        block_min = min(block(:));
        block_max = max(block(:));
        block_width = block_max-block_min;
        if block_width>Twidth % This should never happen! Issue an alert if so.
            beep;
            fprintf('An error occurred: the pixel value range of Block(%d,%d) = [%g, %g]!\n', y, x, block_min, block_max);
        elseif block_max>Tmax
            img(y:y+N-1,x:x+N-1) = block + (Tmax-block_max);
        elseif block_min<Tmin
            img(y:y+N-1,x:x+N-1) = block + (Tmin-block_min);
        % else % if Tmin<=block_min<=block_max<=Tmax (valid range)
        end
    end
end
