function [img_recovered, flow_count] = imrecover_scan(img_dcfree, options)
% [img_recovered, flow_count] = imrecover_scan(img_dcfree, options)
% 
% This function recovers an image from a DC-free image by scanning it with
% a specific order. 
% The main processes involved are Steps II-C/D in Uehara et al.'s paper.
% 
% The input argumen options is a structure variable composed of the
% following fields:
% Tmin, Tmax: minimal/maximal pixel value
% blocksize: size of block
% scan: the scan order (1 = left-upper, 2 = right-upper, 3 = right-bottom,
% 4 = left-bottom)
% dcfree: a value used to resetting the DC coefficient of the first
% reference block
% bound_dc: true/false = the run-time DC-bounding process is on/off
% weighted_block: true/false = an advanced weighting scheme is/isn't
% applied for blocks with two (left/right and bottom/top) references.
% FR_type: the type of flow rate (1 = rate of blocks with
% underflow/overflow, 2 = rate of pixels with underflow/overflow, 3 = rate
% of blocks with underflow/overflow, modulated by the maximal distance from
% the valid range, 4 = rate of sum of underflow/overflow pixel distance
% from the valid range.
% Default values of the above fields: Tmin = 0, Tmax = 255, blocksize = 8,
% bound_dc = true, dcfree = NA, weighted_block = false.
% 
% The return values are the recovered image and the numbers of blocks with
% underflow and overflow when Step II-D is applied.
%
% Copyright (C) 2010 Junaid Jameel Ahmad, Shujun Li (www.hooklee.com)

if nargin<1
    disp('At least one input argument is needed!');
    return;
end

if (~exist('options', 'var') || ~isstruct(options))
    options = []; % Supress warnings when ~isstruct(options)
    options.Tmin = 0;
    options.Tmax = 255;
    options.blocksize = 8;
    options.scan = 1;
    options.bound_dc = true; % This is the default value of our proposed FRM method. For Uehara et al.'s method, it should be false.
    options.weighted_block = false;
    options.FR_type = 1;
else
    % Check if each filed is included in options.
    options_fieldnames = {'Tmin' 'Tmax' 'blocksize' 'scan' 'bound_dc' 'weighted_block' 'FR_type'};
    for i=1:numel(options_fieldnames)
        if ~isfield(options, options_fieldnames{i})
            switch options_fieldnames{i}
                case 'Tmin'
                    options.Tmin = 0;
                case 'Tmax'
                    options.Tmax = 255;
                case 'blocksize'
                    options.blocksize = 8;
                case 'scan'
                    options.scan = 1;
                case 'bound_dc'
                    options.bound_dc = true;
                case 'weighted_block'
                    options.weighted_block = false;
                case 'FR_type'
                    options.FR_type = 1;
            end
        end
    end
end
Tmin = options.Tmin;
Tmax = options.Tmax;
N = options.blocksize;
scan = options.scan;

% Set the DC-free image to have a specific dcfree value.
if isfield(options, 'dcfree')
    img_dcfree = img_dcfree_update(img_dcfree, options.dcfree, options.scan, N, Tmin, Tmax);
end

% Transform img_dcfree to double to avoid potential precision problems.
if isinteger(img_dcfree)
    img_recovered = double(img_dcfree);
else
    img_recovered = img_dcfree;
end

% NOTE: The first dimension of a MATLAB array denotes rows (y-axis) and the
% second denotes columns (x-axis)!
[Y X] = size(img_recovered);
xBlocks = X/N;
yBlocks = Y/N;

switch(scan)
    case 4 % From left-bottom corner
        startX = 0;
        endX = xBlocks-1;
        incX = 1;
        startY = yBlocks-1;
        endY = 0;
        incY = -1;
    case 3 % From right-bottom corner
        startX = xBlocks-1;
        endX = 0;
        incX = -1;
        startY = yBlocks-1;
        endY = 0;
        incY = -1;
    case 2 % From right-top corner
        startX = xBlocks-1;
        endX = 0;
        incX = -1;
        startY = 0;
        endY = yBlocks-1;
        incY = 1;
    otherwise % Default = case 1: From left-top corner
        startX = 0;
        endX = xBlocks-1;
        incX = 1;
        startY = 0;
        endY = yBlocks-1;
        incY = 1;
end

flow_count = 0;
for yb=startY:incY:endY % Rows
    for xb=startX:incX:endX % Columns
        y = (yb*N)+1;
        x = (xb*N)+1;
        adjVal = zeros(1,4);
        neighbour_count = 0;
        % Left reference block.
        if ((scan==1 || scan==4) && xb>0)
            neighbour_count = neighbour_count + 1;
            B = img_recovered(y:y+N-1,x); % The first column of the current block.
            RB = img_recovered(y:y+N-1,x-1); % The first column to the left of the current block.
            adjVal(neighbour_count) = find_best_pattern(RB,B);
        end
        % Right reference block.
        if ((scan==2 || scan==3) && xb<xBlocks-1)
            neighbour_count = neighbour_count + 1;
            B = img_recovered(y:y+N-1,x+N-1); % The last column of the current block.
            RB = img_recovered(y:y+N-1,x+N); % The first column to the right of the current block.
            adjVal(neighbour_count) = find_best_pattern(RB,B);
        end
        % Top reference block.
        if ((scan==1 || scan==2) && yb>0)
            neighbour_count = neighbour_count + 1;
            B = img_recovered(y,x:x+N-1); % The first row of the current block.
            RB = img_recovered(y-1,x:x+N-1); % The first row above the current block.
            adjVal(neighbour_count) = find_best_pattern(RB,B);
        end
        % Bottom reference block.
        if ((scan==3 || scan==4) && yb<yBlocks-1)
            neighbour_count = neighbour_count + 1;
            B = img_recovered(y+N-1,x:x+N-1); % The last row of the current block.
            RB = img_recovered(y+N,x:x+N-1); % The first row below the current block.
            adjVal(neighbour_count) = find_best_pattern(RB,B);
        end
        % Should be only 2 non-zero adjValus, but when there are bugs can
        % be 4.
        if neighbour_count>2
            fprintf('Block (%d:%d,%d:%d): neighbour_count=%d>2 should not happen! Check the code for bugs!\n', ...
                yb,yb+N-1,xb,xb+N-1,neighbour_count);
            pause;
        end
        
        current_block = img_recovered(y:y+N-1,x:x+N-1);
        if (options.weighted_block && neighbour_count>1)
            block = do_weighted_block(current_block, scan, adjVal(1:2));
            if options.bound_dc
                if (min(block(:))<Tmin || max(block(:))>Tmax) % underflow/overflow
                    flow_count = flow_count + 1;
                    block = clip(block);
                end
            end
            img_recovered(y:y+N-1,x:x+N-1) = clip(block); % We can only do this, because bound_dc cannot be easily applied.
        else
            if neighbour_count>0
                adjVal_final = sum(adjVal)/neighbour_count;
            else % neighbour_count == 0 happens for the first block in each scan.
                adjVal_final = 0;
            end
            if options.bound_dc
                adjVal_final0 = adjVal_final;
                adjVal_final = bound_DC(current_block, adjVal_final, Tmin, Tmax);
                if adjVal_final0~=adjVal_final % underflow/overflow
                    switch options.FR_type
                        case 2 % Number of pixels with underflow/overflow as the flow rate.
                            current_block2 = current_block(:) + adjVal_final0;
                            flow_count = flow_count + sum(current_block2>Tmax | current_block2<Tmin);
                        case 3 % Maximal overflow/underflow pixel value as the flow rate.
                            current_block2 = current_block(:) + adjVal_final0;
                            overflow_max = max(max(current_block2-Tmax),0);
                            underflow_max = max(max(Tmin-current_block2),0);
                            flow_count = flow_count + max(overflow_max,underflow_max);
                        case 4 % Sum of pixel overflow/underflow values as flow rate.
                            current_block2 = current_block(:) + adjVal_final0;
                            flow_count = flow_count + sum(current_block2(current_block2>Tmax)-Tmax) ...
                                + sum(Tmin-current_block2(current_block2<Tmin));
                        otherwise % Default: Number of blocks with underflow/overflow as the flow rate. 
                            flow_count = flow_count + 1;
                    end
                end
            end
            current_block = current_block + adjVal_final;
            if ~options.bound_dc
                if (any(current_block(:)<Tmin) || any(current_block(:)>Tmax))
                    flow_count = flow_count + 1; % Underflow/Overflow
                end
            end
            img_recovered(y:y+N-1,x:x+N-1) = current_block;
        end
    end
end


function [adjVal,minMSD,pattern] = find_best_pattern(RB,B)
% [adjVal,minMSD,pattern] = find_best_pattern(RB,B)
% This function finds the smoothest patterns among the three ones
% defined in the paper.
% RB, B are row or column vectors extracted from reference and current
% blocks, respectively.

% Three patterns.
RBs={RB,RB(2:end),RB(1:end-1)};
Bs={B,B(1:end-1),B(2:end)};

% Calculate the means and remove them from the three patterns.
RM = zeros(1,3);
M = zeros(1,3);
MSD = zeros(1,3);
for i=1:3
    RM(i) = mean(RBs{i});
    M(i) = mean(Bs{i});
    RBs{i} = RBs{i} - RM(i);
    Bs{i} = Bs{i} - M(i);
    MSD(i) = mean((RBs{i}-Bs{i}).^2);
end

% Find the minimal MSD and the corresponding pattern index and adjusted
% intentity for the current block.
[MSD_min,MSD_min_index] = min(MSD);
minMSD = MSD_min(1);
pattern = MSD_min_index(1);
adjVal = (RM(pattern)-M(pattern));

function y = do_weighted_block(x, scan, adjVal12)
% This function performs the weigted average for a block x based on its 
% distance from the diagonal.
% adjVal12 is the adjustment values of two neighboring blocks. 
% y is the adjusted block after weighted average has been applied to all
% pixels.

if(scan==2 || scan==4)
    adjVal12 = adjVal12([2 1]);
end

N = size(x,1);
N_1 = N-1;

x = rot90(x,(scan-1));
z = spdiags(x,-N_1:N_1);
[rows cols] = size(z); % cols = N_1*2+1, rows = N;
for i=1:cols
    w2 = (i-1)/(cols-1);
    W12 = (1 - w2)*adjVal12(1) + w2*adjVal12(2);
    
    v = z(:,i);
    if(i <= (N_1+1)) 
        t0 = 1; tEnd = i;
    else
        t0 = i-N+1; tEnd = rows;
    end
    for j=t0:tEnd
        v(j) = v(j) + W12;
    end
    z(:,i) = v;
end
y = full(spdiags(z,-N_1:N_1, N, N));
y = rot90(y,-(scan-1));
